// operators and operations on matrices

namespace mathlib {

#define MATHLIB_MATRIX_OPERATOR_FLOAT(name, op) \
	template<class A, int R, int C> inline \
	const tmatrix<matrix_binop<binop_##name, A, float_keeper> > \
	op(const tmatrix_dim<R, C, A> &a, float b) \
		{ return tmatrix<matrix_binop<binop_##name, A, \
			float_keeper> >(a, b); }

#define MATHLIB_FLOAT_OPERATOR_MATRIX(name, op) \
	template<class A, int R, int C> inline \
	const tmatrix<matrix_binop<binop_##name, float_keeper, A> > \
	op(float a, const tmatrix_dim<R, C, A> &b) \
		{ return tmatrix<matrix_binop<binop_##name, \
			float_keeper, A> >(a, b); }

#define MATHLIB_MATRIX_UNARY_OPERATOR(name, op) \
	template<class A, int R, int C> inline \
	const tmatrix<matrix_unop<unop_##name, A> > \
	op(const tmatrix_dim<R, C, A> &a) \
		{ return tmatrix<matrix_unop<unop_##name, A> >(a); }

#define MATHLIB_MATRIX_OPERATOR_MATRIX_ADD(name, op) \
	template<class A, class B, int R, int C> inline \
	const tmatrix<matrix_binop<binop_##name, A, B> > \
	op(const tmatrix_dim<R, C, A> &a, const tmatrix_dim<R, C, B> &b) \
		{ return tmatrix<matrix_binop<binop_##name, A, B> >(a, b); }

#define MATHLIB_VECTOR_OPERATOR_VECTOR(name, op) \
	template<class A, class B, int R> inline \
	const tmatrix<matrix_binop<binop_##name, A, B> > \
	op(const tmatrix_dim<R, 1, A> &a, const tmatrix_dim<R, 1, B> &b) \
		{ return tmatrix<matrix_binop<binop_##name, A, B> >(a, b); }

// matrix multiplication
template<class A, class B, int C, int D, int E> inline
const tmatrix<matrix_binop<binop_mmul<D>, A, B> >
operator*(const tmatrix_dim<C, D, A> &a, const tmatrix_dim<D, E, B> &b)
	{ return tmatrix<matrix_binop<binop_mmul<D>, A, B> >(a, b); }

MATHLIB_FLOAT_OPERATOR_MATRIX(mul, operator*)
MATHLIB_MATRIX_OPERATOR_FLOAT(mul, operator*)
MATHLIB_MATRIX_OPERATOR_MATRIX_ADD(add, operator+)
MATHLIB_MATRIX_OPERATOR_MATRIX_ADD(sub, operator-)
MATHLIB_MATRIX_OPERATOR_FLOAT(div, operator/)
MATHLIB_VECTOR_OPERATOR_VECTOR(mul, operator*)
MATHLIB_VECTOR_OPERATOR_VECTOR(div, operator/)
MATHLIB_MATRIX_UNARY_OPERATOR(negate, operator-)

// vector product
template<class A, class B> inline
const tmatrix<matrix_binop<binop_cross, A, B> >
cross(const tmatrix_dim<3, 1, A> &a, const tmatrix_dim<3, 1, B> &b)
	{ return tmatrix<matrix_binop<binop_cross, A, B> >(a, b); }

// scalar product
template<class A, class B, int R> inline
const tmatrix<matrix_binop<binop_dot<R>, A, B> >
dot(const tmatrix_dim<R, 1, A> &a, const tmatrix_dim<R, 1, B> &b)
	{ return tmatrix<matrix_binop<binop_dot<R>, A, B> >(a, b); }

// transpose
template<class A, int R, int C> inline
const tmatrix<matrix_unop<unop_cindex<index_transpose>, A> >
transpose(const tmatrix_dim<R, C, A> &a)
	{ return tmatrix<matrix_unop<unop_cindex<index_transpose>, A> >(a); }

// make a translation matrix
template<class A> inline
const tmatrix<matrix_unop<unop_translation, A> >
translation(const tmatrix_dim<3, 1, A> &a)
	{ return tmatrix<matrix_unop<unop_translation, A> >(a); }

// make a scaling matrix
template<class A> inline
const tmatrix<matrix_unop<unop_scaling, A> >
scaling(const tmatrix_dim<3, 1, A> &a)
	{ return tmatrix<matrix_unop<unop_scaling, A> >(a); }
inline const tmatrix<matrix_unop<unop_scaling, float_keeper> > scaling(float a)
	{ return tmatrix<matrix_unop<unop_scaling, float_keeper> >(a); }

// make a rotation matrix
template<class A> inline
const tmatrix<matrix_binop<binop_rotation, A, float_keeper> >
rotation(const tmatrix_dim<3, 1, A> &axis, float radians) {
	return tmatrix<matrix_binop<binop_rotation, A, float_keeper> >
		(axis, radians);
}

// make a perspective projection matrix
const matrix<4> frustum(float left, float right, float bottom, float top,
	float near, float far);

// make an orthographic projection matrix
const matrix<4> ortho(float left, float right, float bottom, float top,
	float near, float far);

// inverse
template<class A, int RC> inline
const tmatrix<matrix_unop<unop_inverse, A> >
inverse(const tmatrix_dim<RC, RC, A> &a)
	{ return tmatrix<matrix_unop<unop_inverse, A> >(a); }

// adjoint
template<class A, int R, int C> inline
const tmatrix<matrix_unop<unop_adjoint, A> >
adjoint(const tmatrix_dim<R, C, A> &a)
	{ return tmatrix<matrix_unop<unop_adjoint, A> >(a); }

// perspective division
template<class A> inline
const tmatrix<matrix_unop<unop_wdiv, A> > wdiv(const tmatrix_dim<4, 1, A> &a)
	{ return tmatrix<matrix_unop<unop_wdiv, A> >(a); }

// normalize vector
template<int R, class A> inline
const tmatrix<matrix_unop<unop_normalize, A> >
norm(const tmatrix_dim<R, 1, A> &a)
	{ return tmatrix<matrix_unop<unop_normalize, A> >(a); }

// reciprocal of vector
template<int R, class A> inline
const tmatrix<matrix_unop<unop_reciprocal, A> >
reciprocal(const tmatrix_dim<R, 1, A> &a)
	{ return tmatrix<matrix_unop<unop_reciprocal, A> >(a); }

// expand vector
template<class A, int R> inline
const tmatrix<matrix_binop<binop_expand, A, float_keeper> >
expand(const tmatrix_dim<R, 1, A> &a, float b)
	{ return tmatrix<matrix_binop<binop_expand, A, float_keeper> >(a, b); }

// remove row R and column C
template<int R, int C, class A, int D, int E> inline
const tmatrix<matrix_unop<unop_cindex<index_cofactor<R, C> >, A> >
cofactor(const tmatrix_dim<D, E, A> &a) {
	return tmatrix<matrix_unop<unop_cindex<index_cofactor<R, C> >, A> >(a);
}

// same as above, usable from impementations of other operations
template<int R, int C, class T> inline
const tmatrix<matrix_unop<unop_cindex<index_cofactor<R, C> >, T> >
cofactor_bm(const T &a) {
	return tmatrix<matrix_unop<unop_cindex<index_cofactor<R, C> >, T> >(a);
}

// evaluate to temporary
template<int R, int C, class T> inline
const matrix<R, C> temp(const tmatrix_dim<R, C, T> &a) { return a; }

// determinant
inline float det(float a1, float a2, float b1, float b2)
	{ return a1*b2 - a2*b1; }
inline float det(basic_matrix<2, 1> a, basic_matrix<2, 1> b)
	{ return a.x*b.y - a.y*b.x; }
inline float det_bm(const basic_matrix<1, 1> &m) { return m(0,0); }
inline float det_bm(const basic_matrix<2, 2> &m)
	{ return m(0,0)*m(1,1) - m(1,0)*m(0,1); }
inline float det(float a1, float a2, float a3, float b1, float b2, float b3, 
	float c1, float c2, float c3)
{
	return a1*b2*c3 + a2*b3*c1 + a3*b1*c2 - a3*b2*c1 - a1*b3*c2 - a2*b1*c3;
}
inline float det(basic_matrix<3, 1> a, basic_matrix<3, 1> b,
	basic_matrix<3, 1> c)
{
	return det(a.x, a.y, a.z, b.x, b.y, b.z, c.x, c.y, c.z);
}
float det_bm(const basic_matrix<3, 3> &);
float det_bm(const basic_matrix<4, 4> &);

// dispatch determinant
template<int A> struct select_determinant {
	static float det(const basic_matrix<A, A> &m)
		{ return det_bm(m); }
};
template<int RC, class A> inline float det(const tmatrix_dim<RC, RC, A> &a)
	{ return select_determinant<RC>::det(a); }

// squared length of vector
inline float abs_squared_bm(const bm21 &a)
	{ return a.x*a.x + a.y*a.y; }
inline float abs_squared_bm(const bm31 &a)
	{ return a.x*a.x + a.y*a.y + a.z*a.z; }
inline float abs_squared_bm(const bm41 &a)
	{ return a.x*a.x + a.y*a.y + a.z*a.z + a.w*a.w; }

// dispatch abs_squared
template<int A> struct select_abs_squared {
	static inline float asq(const basic_matrix<A, 1> &a)
		{ return abs_squared_bm(a); }
};
template<int R, class A> inline
float abs_squared(const tmatrix_dim<R, 1, A> &a)
	{ return select_abs_squared<R>::asq(a); }

// length of vector
template<int R, class A> inline
float abs(const tmatrix_dim<R, 1, A> &a)
	{ return std::sqrt(abs_squared(a)); }

// sum of elements
inline float sum_bm(const bm21 &a)
	{ return a.x + a.y; }
inline float sum_bm(const bm31 &a)
	{ return a.x + a.y + a.z; }
inline float sum_bm(const bm41 &a)
	{ return a.x + a.y + a.z + a.w; }

// dispatch sum
template<int A> struct select_sum {
	static inline float sum(const basic_matrix<A, 1> &a)
		{ return sum_bm(a); }
};
template<int R, class A> inline
float sum(const tmatrix_dim<R, 1, A> &a)
	{ return select_sum<R>::sum(a); }


// signed area of triangle (ccw)
inline float area(const vector<2> &a, const vector<2> &b, const vector<2> &c)
	{ return det(b-a, c-a)/2; }
// area of triangle
inline float area(const vector<3> &a, const vector<3> &b, const vector<3> &c)
	{ return abs(cross(b-a, c-a))/2; }

// signed area of quadriliteral (ccw)
inline float area(const vector<2> &a, const vector<2> &b,
	const vector<2> &c, const vector<2> &d) {
	vector<2> ac = c-a;
	return (det(b-a, ac)+det(ac, d-a))/2;
}
// area of quadriliteral
inline float area(const vector<3> &a, const vector<3> &b,
	const vector<3> &c, const vector<3> &d) {
	vector<3> ac = c-a;
	return (abs(cross(b-a, ac))+abs(cross(ac, d-a)))/2;
}

template<int R, int C, int I, class T>
struct matrix_assigner {
	T &dest;
	matrix_assigner(T &d) : dest(d) {}
	matrix_assigner<R, C, I+1, T> operator,(float f)
		{ dest.template element<I/C, I%C>() = f; return dest; }
};

// assign to matrix
template<int R, int C, class T> inline
matrix_assigner<R, C, 1, T> operator<<(tmatrix_dim<R, C, T> &dest, float f)
	{ dest.template element<0,0>() = f; return dest; }

// assign to matrix
template<int R, int C, class T> inline
matrix_assigner<R, C, 1, T> operator<<(const tmatrix_dim<R, C, T> &d, float f) {
	T &dest = const_cast<T &>(static_cast<const T &>(d));
	dest.template element<0,0>() = f; return dest;
}

// convert to vector
template<class T> vector<2> vector_xy(T v) { return vector<2>(v.x, v.y); }
template<class T> vector<3> vector_xyz(const T &v)
	{ return vector<3>(v.x, v.y, v.z); }
template<class T> vector<4> vector_xyzw(const T &v)
	{ return vector<4>(v.x, v.y, v.z, v.w); }

template<class T> vector<3> vector_rgb(const T &v)
	{ return vector<3>(v.r, v.g, v.b); }
template<class T> vector<4> vector_rgba(const T &v)
	{ return vector<4>(v.r, v.g, v.b, v.a); }

template<class T> vector<2> vector_st(T v) { return vector<2>(v.s, v.t); }
template<class T> vector<3> vector_stu(const T &v)
	{ return vector<3>(v.s, v.t, v.u); }
template<class T> vector<4> vector_stuv(const T &v)
	{ return vector<4>(v.s, v.t, v.u, v.v); }

// compare matrices
template<int A, int B>
bool operator==(const matrix<A, B> &a, const matrix<A, B> &b)
{
	for (int i=0; i!=A*B; ++i)
		if (a.data()[i] != b.data()[i])
			return false;
	return true;
}
template<int A, int B> inline
bool operator!=(const matrix<A, B> &a, const matrix<A, B> &b)
	{ return !(a == b); }

// compare vectors
template<int A>
bool operator==(const vector<A> &a, const vector<A> &b)
{
	for (int i=0; i!=A; ++i)
		if (a.data()[i] != b.data()[i])
			return false;
	return true;
}
template<int A> inline
bool operator!=(const vector<A> &a, const vector<A> &b)
	{ return !(a == b); }

// print matrix
template<int A, int B>
std::ostream &operator<<(std::ostream &s, const matrix<A, B> &m)
{
	s << "\n(";
	for (int r=0; r<A-1; ++r) {
		s << '(';
		for (int c=0; c<B-1; ++c)
			s << m(r,c) << ',';
		s << m(r,B-1) << ")\n ";
	}
	s << '(';
	for (int c=0; c<B-1; ++c)
		s << m(A-1,c) << ',';
	return s << m(A-1,B-1) << "))";
}

// print vector
template<int A>
std::ostream &operator<<(std::ostream &s, const vector<A> &m)
{
	s << "(";
	for (int i=0; i<A-1; ++i)
		s << m(i) << ',';
	return s << m(A-1) << ")t";
}

} // namespace mathlib
